--[[--------------------------------------------------------------------------------------
This is a scan hierarchy class system implemented using niftyreg tools

Most methods are not specific to the [Conquest](http://ingenium.home.xs4all.nl/dicom.html) DICOM server and may be run in any lua interpreter
[Niftyreg](https://sourceforge.net/projects/niftyreg/) tools including checkcudacard are expected to be present as defined below.
As viewer [MRIcron](http://www.mccauslandcenter.sc.edu/mricro/mricron/index.html) is used, for now only for methods _scan:view()_ and _scan:viewdef()_. 
To create a GUI [IUP](http://www.tecgraf.puc-rio.br/iup/) is used, for now only for method _scan:plot()_. [7zips 7za](http://www.7-zip.org/) is used to
unzip gzipped nifty files in method _scan:read nifty()_.

#scan: lua object representing a scan and its transformations
  :`new(filename)`       
    Create scan representing nifty file (`filename` optional, uses tempfile if none passed)
    The scan has embedded datafiles for rigid and deformable registration and fusion
  :`free()`
    Deletes nifty files representing scan and its registrations
  :`assign(a)`
    Copy nifty files representing scan and its registrations from scan `a`
  :`read_dicom(server, q)`
    Read series from `server` based on dicom query object `q` or string 'PatientID:SeriesInstanceUID'
    or table q, where q[1] is a string in format 'PatientID:SeriesInstanceUID'.
    The scan is read with correct DICOM to Nifty transformations, e.g., axial and coronal series
    from the same study will remain inherently matched. This routine is Conquest DICOM server
    specific
  :`write_dicom(usecoords, replacevalues)`
    Save series in local dicom server (using the dicomserver data/incoming folder that must exist)
    This routine can only be called after `scan:read_dicom` as DICOM headers are reused. if `usecoords` 
    is set, ImagePositionPatient and ImageOrientationPatient are replaced from the Nifty headers.
    The `replacevalues` table is used to overrule DICOM header e.g., {SeriesDescription = 'ReOriented'}
    This routine is Conquest DICOM server specific
  :`read_nifty(filename)`
    Copies nifty scan `filename` a into scan object (will also uncompress gz file)
    This routine expects 7za present (see `_7za` config parameter) if gz file passed
  :`write_nifty(filename)`
    Copy embedded nifty1.1 file to `filename` for external use
  :`isempty()`
    Returns true if scan file is not present (e.g., just after **scan** creation)
  :`as(scan)`
    returns **self** resampled to the coordinates of `scan` searching registrations within 
    the **container** the scan is
  :`add(scan)` :`sub(scan)` :`mul(scan)` :`div(scan)`
    Add/sub/mul/div `scan` to **self**, or Add/sub/mul/div **self** with a constant.
    Note that scans to be processed may differ in orientation and resolution
    The input `scan` is resampled using _scan:as()_ to the output **self**. If self 
    `isempty()` is true, `scan` is just copied into **self**
  :`thr(constant)`
    Threshold scan with `constant`
  :`smoothGaussian(a, b, c)`
    Smooth scan with Gaussian kernel `a, b, c`
  :`smooth(a)`
    Smooth scan with B-spline kernel `a`
  :`reg_aladin(a, pars)`
    Rigid match scan using `a` as reference, modifies **self.aff** transform. 
    If `a` is a string or number, it is used as self.owner[a], i.e., to name scan in the same **container**
    If `pars` is set, these are passed as command line parameters to reg_aladin
  :`reg_f3d(a)`
    Deformable match scan usign `a` as reference, modifies **self.dvf** and **self.invdvf**
    If `a` is a string or number, it is used as self.owner[a], i.e., to name scan in the same **container**
  :`view()`
    Uses external viewer (mricron) to show the scan in Nifty1.1 format
  :`viewdef()`
    Uses external viewer (mricron) to show the scan's DVF in Nifty1.1 format
  :`append(a)`
    Creates a 4D scan in **self** by appending scan `a`, assumes **self** and `a` have same dimensions
  :`slice(frame)`
    Slices scan **self** at one timeframe
  :`slicefrom(a, frame)`
    Slices scan `a` into **self** at one timeframe
  `:transform(tx, ty, tz, rx, ry, rz)` transforms the scan by updating the embedded SForm matrix
  

#container: overloaded generic lua table
  `:new()`
    Creates empty **container**, address e.g., using _self.name_ or _self[index]_
  `:free()`
    Free container and all its children (e.g. also frees scans)
  `:readfiles(dir, mat, f, max)`
    Fill **self** from directory `dir`, reading `max` files (default all) matching name `mat` 
    with function `f(filename)` returning a scan
  `:write(file)`
    Create lua script in file that when run will restore contents of container
  `:read(file)`
    Restores contents of container from lua script written with _container:write(file)_
  `:plot(function, caption)`
    Plot `function`(self[i]) against i with `caption` {window, plot, x_axis, y_axis}, 
    `function` may be multi-valued. Note: require IUP and CD installed in clibs folder
  `:read_dicom(serverAE, q, f, max, split4d)`
    Fill **self** from DICOM server, where q is a DICOM query object returning a list of 
    series, or q is a table of patid:seriesuid strings, or a q is a table of query 
    objects or a DicomArray of query objects; default function f(server, q) reads all series
    with more than 9 slices; if `split4d` is set, 4D series are split into separate scans
    
todo: add :addscan(name, folder)
   
-----------------------------------------------------------------------------------------
]]--

-- Updates:
-- mvh 20130215: Added possibility to write nifty file for each slice
-- mvh 20130215: Fixed output transform ;->>>
-- mvh 20130217: Accept table for filenames
-- mvh 20130225: Added option to replicate slices (if each slice to become a volume)
-- mvh 20130305: Added option to use nifty s-form coordinates in reread_nifty
-- mvh 20130305: Improved write_nifty precision using first and last slice to get direction
-- mvh 20130305: Use replacevalues table to overrule dicom tags in reread_nifty
-- mvh 20130305: Added scan class for simple registration etc in dicom server
-- mvh 20130314: Implemented reg_aladin, reg_f3d etc
-- mvh 20130318: Added container, container:reg_f3d etc
-- mvh 20130319: Added container:contains, container:path, container:readfiles, scan:read_nifty
-- mvh 20130320: Use system() call to suppress messages; added container:read/write
-- mvh 20130320: Added container:process and container:plot
-- mvh 20130321: Added GPU check
-- mvh 20130322: Set package.cpath from container:plot; added scan:assign, append_nifty, scan:append
-- mvh 20130326: Split of niftyutil and niftydicom
-- mvh 20130409: Moved directory/file configs forwards
-- mvh 20130416: Added container:read_dicom; function f may return nil here and in container:readfiles
-- mvh 20130417: Added invaff to scan; updated path to show direct transforms
-- mvh 20130430: Removed leak is scan:as(); make :as() always return at least a copy
-- mvh 20130619: linux work, use packages niftyreg, mricron, p7zip.full
-- mvh 20130628: Sort dicom series on time before reading container; sort images on number
-- mvh 20130628: Put info table into scan object; print some dicom progress for debugging
-- mvh 20130701: Sort series same time on series number; added slice, slicefrom, split4d, dimensions
-- mvh 20130701: Assign copies dicomfiles and info deep
-- mvh 20130702: Fixed recursive seeking in normal container; added ref_rigid using NKI's LuaMatch scriptable app
-- mvh 20130703: Container:read_dicom reads scans with >=10 slices; scan:read_dicom stores extra info in dicomfiles
-- mvh 20130704: downsize sets affine to unity to force transform if needed; added scan:normalize 
--               using NKI's LuaMatch
-- mvh 20130705: name items in dicomfiles and query table for readability; 
--               let dicom reader set filename of nifty files
-- mvh 20130708: reject invalid 4D data; todo make folders for containers; 
--		 use ImageType if SeriesDescription is empty; reject DERIVED data
-- mvh 20130711: Added info to container
-- mvh 20130715: Delete scan passed to scan:new if it exists
-- mvh 20130715: scan:_resample will now use updSform if possible; added scan:transform method
-- mvh 20130716: added scan:reg_niftkAffine and associated niftk folder; it has a pesky - on row/col 2 ...
-- mvh 20140319: reduced release for with 1.4.17d; removed reference to unreleased tools


-- configuration of folders and external executables
local niftyreg  = "c:\\dicomserver1417d\\"
local niftk     = "C:\\Program Files (x86)\\NifTK 12.11\\bin\\"
local viewerexe = "c:\\Program Files (x86)\\mricron\\mricron.exe"
local _7za      = "c:\\dicomserver1417d\\7za.exe"
local tmpdir    = "c:\\dicomserver1417d\\niftidata\\"
local checkcudacard = "checkcudacard.exe"
local exe       = '.exe'
local slashb    = ' /b'

if string.sub(os.getenv('HOME') or '', 1, 1)=='/' then
  niftyreg      = ""
  niftk         = ""
  viewerexe     = "mricron"
  _7za          = "7za"
  tmpdir        = "/tmp/"
  checkcudacard = "checkcudacard"
  exe           = ''
  slashb        = ' -1'
end

require('niftyutil')

function tempfile()
  if string.sub(os.getenv('HOME') or '', 1, 1)=='/' then
    return os.tmpname()
  else
    return string.sub(tmpdir, 1, #tmpdir-1)..os.tmpname()
  end
end

-- Test if GPU is available
local gpu = ''
if os.execute('"'..niftyreg..checkcudacard..'"')==0 then
  print('cuda card detected')
  gpu = '-gpu '
  for line in io.popen('"'..niftyreg..checkcudacard..'"'):lines() do
    print('compute capability', line)
  end
else
  print('No cuda card detected')
end

--/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
-- scan object
--/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

scan = {}
scan.__index = scan

function scan:new(f)
  local s = {}
  setmetatable(s,scan)
  s.classname  = 'scan'
  s.filename   = f or tempfile() .. '.nii'
  if fileexists(s.filename) then os.remove(s.filename) end
  s.dicomfiles = {}
  s.info       = {}
  s.aff        = ''
  s.invaff     = ''
  s.cpp        = ''
  s.invdvf     = ''
  s.dvf        = ''
  --s.name     = '' -- set by container
  --s.owner    = {} -- set by container
  return s
end

function scan:free()
  if self.filename~='' then os.remove(self.filename) end
  if self.aff     ~='' then os.remove(self.aff) end
  if self.invaff  ~='' then os.remove(self.invaff) end
  if self.dvf     ~='' then os.remove(self.dvf) end
  if self.invdvf  ~='' then os.remove(self.invdvf) end
  if self.cpp     ~='' then os.remove(self.cpp) end
end

function scan:assign(a)
  filecopy(a.filename, self.filename)
  if a.aff~='' then 
    self.aff = self.filename .. '.aff.txt'
    filecopy(a.aff, self.aff) 
  end
  if a.invaff~='' then 
    self.invaff = self.filename .. '.invaff.txt'
    filecopy(a.invaff, self.invaff) 
  end
  if a.dvf~='' then 
    self.dvf = self.filename .. '.def'
    filecopy(a.dvf, self.dvf) 
  end
  if a.invdvf~='' then 
    self.invdvf = self.filename .. '.invdef'
    filecopy(a.invdvf, self.invdvf) 
  end
  if a.cpp~='' then 
    self.cpp = self.filename .. '.cpp'
    filecopy(a.cpp, self.cpp) 
  end
  for k,v in pairs(a.dicomfiles) do
    self.dicomfiles[k] = a.dicomfiles[k]
  end
  for k,v in pairs(a.info) do
    self.info[k]      = a.info[k]
  end
end

-- q can be PatientID:SeriesInstanceUID or query DICOM object
-- this function only works in a DICOM server environment
function scan:read_dicom(server, q)
  require('niftydicom')
  
  if type(q)=='string' then
    local r = q;
    q = DicomObject:new()
    q.PatientID         = string.gsub(r, ':.*', '')
    q.SeriesInstanceUID = string.gsub(r, '.*:', '')
  elseif type(q)=='table' then
    local r = q;
    q = DicomObject:new()
    q.PatientID         = string.gsub(r[1] or r.name, ':.*', '')
    q.SeriesInstanceUID = string.gsub(r[1] or r.name, '.*:', '')
  end
  q.QueryRetrieveLevel = 'IMAGE'
  if q.SOPInstanceUID ==nil then q.SOPInstanceUID  = '' end
  if q.InstanceNumber ==nil then q.InstanceNumber  = '' end
  if q.AcquisitionTime==nil then q.AcquisitionTime = '' end
  if q.ImageType      ==nil then q.ImageType       = '' end
  if q.SliceLocation  ==nil then q.SliceLocation   = '' end
  
  local r = dicomquery(server, 'IMAGE', q)
  self.dicomfiles = {}
  for i=0, #r-1 do
    table.insert(self.dicomfiles, 
      {name            = r[i].PatientID..':'..r[i].SOPInstanceUID, 
       InstanceNumber  = r[i].InstanceNumber  or '', 
       AcquisitionTime = r[i].AcquisitionTime or '', 
       ImageType       = r[i].ImageType       or '', 
       SliceLocation   = r[i].SliceLocation   or ''})
  end
  table.sort(self.dicomfiles, function(a, b) 
                                return (tonumber(a.InstanceNumber) < tonumber(b.InstanceNumber)) 
			      end)
    
  write_nifty(self.dicomfiles, self.filename)
end

-- create new series on DICOM server
-- this function only works in a DICOM server environment
function scan:write_dicom(a, b)
  require('niftydicom')
  reread_nifty(self.dicomfiles, self.filename, a, b)
end

-- read nifty or nifty.gz into scan object
function scan:read_nifty(name)
  if string.find(name, '%.gz')~=nil then
    local newname;
    newname = string.gsub(name, ".*\\", "")
    newname = string.gsub(name, ".*/", "")
    newname = tmpdir..string.gsub(newname, "%.gz", "")
    os.remove(newname)
    osexecute('"'.._7za..'" e -o'..tmpdir .. ' ' .. name)
    filecopy(newname, self.filename)
    os.remove(newname)
  else
    filecopy(name, self.filename)
  end
end

-- write scan as external nifty; todo add gz option
function scan:write_nifty(name)
  filecopy(self.filename, name)
end

function scan:isempty()
  return not fileexists(self.filename)
end

-- returns nifti imagedimensions structure as table
function scan:dimensions()
  require "pack"

  if not fileexists(self.filename) then
    return
  end
  
  function pack(...)
    return arg
  end
  
  local g = io.open(self.filename, 'rb')
  if g==nil then
    error('scan:dimensions: cannot read file '..self.filename)
  end
  local header = g:read(352) -- read some parts of nifty header only
  local headerkey        = pack(string.unpack(header, 'iA10A18ihA1A1'))
  local imagedimensions  = pack(string.unpack(header, 'hhhhhhhh hhhhhhh hhh ffffffff ffffffff ii', headerkey[1]));
  table.remove(imagedimensions,  1)
  g:close()
  
  return imagedimensions
end

-- read rigid match as table
function scan:rigidmatch()
  require('niftyutil')
  if self.invaff=='' or not fileexists(self.invaff) then
    return
  end
  
  local s = {}
  for line in io.lines(self.invaff) do 
    table.insert(s, split(line, ' '))
  end
  return s
end

function scan:_resample(step, final)
  local r = scan:new()
  local s
  if step.aff~='' and step.aff~=nil and not final then
    s = '"'..niftyreg..'reg_transform'..exe..'" -ref '..step.target..' -updSform '..self.filename..' '..step.aff..' '..r.filename
    osexecute(s)
  elseif step.aff~='' and step.aff~=nil and final then
    s = '"'..niftyreg..'reg_resample'..exe..'" -ref '..step.target..' -flo '..self.filename..' -res '..r.filename
    s = s .. ' -aff ' .. step.aff
    osexecute(s)
  elseif step.invdvf~='' and step.invdvf~=nil then
    s = '"'..niftyreg..'reg_resample'..exe..'" -ref '..step.target..' -flo '..self.filename..' -res '..r.filename
    s = s .. ' -def ' .. step.invdvf
    osexecute(s)
  elseif not final then
    filecopy(step.source, r.filename)
  else
    s = '"'..niftyreg..'reg_resample'..exe..'" -ref '..step.target..' -flo '..self.filename..' -res '..r.filename
    osexecute(s)
  end
  return r
end

-- todo: implement to avoid multiple resamples
function scan:_updatetransform(step)
end  

-- resample self in coordinates of a
function scan:as(_a)
  local r = self
  local a, b = self:findconnection(_a)
  if a==nil then 
    error((self:fullname() or '')..':as('..(_a:fullname() or '') .. ') - no relation found between scans')
  end
  for i=1,#b do
    if (b[i].aff~='') or (b[i].invdvf~='') or (i==#b) then
      local a=r
      r = r:_resample(b[i], i==#b)
      if a~=self then a:free() end
    end
  end
  if r==self then
    r = scan:new()
    r:assign(self)
  end
  return r
end
  
-- calculations on self (internal function)
function scan:op(op, a)
  if type(a)=='table' then
    if self:isempty() then
      self:assign(a)
    else
      local r = a:as(self)
      osexecute('"'..niftyreg..'reg_tools'..exe..'" -in '..self.filename..' -out '..self.filename..' -'..op..' '..r.filename)
      r:free()
    end
  else
    osexecute('"'..niftyreg..'reg_tools'..exe..'" -in '..self.filename..' -out '..self.filename..' -'..op..' '..tostring(a))
  end
end
  
function scan:add(a)
  self:op('add', a)
end

function scan:sub(a)
  self:op('sub', a)
end

function scan:mul(a)
  self:op('mul', a)
end

function scan:div(a)
  self:op('div', a)
end

function scan:thr(a)
  self:op('thr', a)
end

function scan:rms(a)
  local r = a:as(self)
  for line in io.popen('"'..niftyreg..'reg_tools'..exe..'" -in '..self.filename..' -out '..self.filename..' -rms '..r.filename):lines() do
    r:free()
    return line + 0
  end
end

function scan:smooth(a)
  osexecute('"'..niftyreg..'reg_tools'..exe..'" -in '..self.filename..' -out '..self.filename..' -smo ' .. tostring(a))
end

function scan:smoothGaussian(a, b, c)
  osexecute('"'..niftyreg..'reg_tools'..exe..'" -in '..self.filename..' -out '..self.filename..' -smoG ' .. a .. ' ' .. b .. ' ' .. c)
end

function scan:downsize()
  osexecute('"'..niftyreg..'reg_tools'..exe..'" -in '..self.filename..' -out '..self.filename..' -down')
  
  if self.aff~='' or self.invdvf~='' then
    return
  end
  
  self.aff    = self.filename .. '.aff.txt'
  self.invaff = self.filename .. '.invaff.txt'
  local f = io.open(self.invaff, 'wt')
  f:write('1 0 0 0\n')
  f:write('0 1 0 0\n')
  f:write('0 0 1 0\n')
  f:write('0 0 0 1\n')
  f:close()
  filecopy(self.invaff, self.aff)
end

function scan:prepareforreg(a_)
  local target=self
  if type(a_)=='string' or type(a_)=='number' then
    local save_a = a_
    a_ = nil
    if self.owner then a_=self.owner[save_a] end
  end
  if a_==nil  then error 'reference scan not defined'  end
  if a_==self then error 'cannot match scan to itself' end
  
  return a_, target
end

function scan:reg_aladin(_a, pars)
  local a, target = self:prepareforreg(_a)
   
  if target.aff=='' then target.aff = target.filename .. '.aff.txt' end
  if target.invaff=='' then target.invaff = target.filename .. '.invaff.txt' end
  pars = pars or ''
  
  -- -ref == -target -flo == -source -nac -rigOnly -noSym
  osexecute('"'..niftyreg..'reg_aladin'..exe..'" -ref '..a.filename..' -flo ' .. self.filename .. ' -aff ' .. target.aff .. ' ' .. pars)
  osexecute('"'..niftyreg..'reg_transform'..exe..'" -ref '..a.filename..' -invAffine ' .. target.aff .. '  ' .. target.invaff)
  if target.dvf~='' then 
    os.remove(target.dvf)
    target.dvf = ''
  end  
  if target.invdvf~='' then 
    os.remove(target.invdvf)
    target.invdvf = ''
  end  
  if target.cpp~='' then 
    os.remove(target.cpp)
    target.cpp = ''
  end
end

function scan:transform(tx, ty, tz, rx, ry, rz)
  local t = tempfile()
  tx = tx or 0
  ty = ty or 0
  tz = tz or 0
  rx = rx or 0
  ry = ry or 0
  rz = rz or 0
  osexecute('"'..niftyreg..'reg_transform'..exe..'" -ref '..self.filename..' -makeAffine ' .. 
    rx .. ' ' .. ry .. ' ' .. rz .. ' ' .. tx .. ' ' .. ty .. ' ' .. tz .. 
    ' 1 1 1 0 0 0 ' .. t)
  s = '"'..niftyreg..'reg_transform'..exe..'" -ref '..self.filename..
    ' -updSform '..self.filename..' '.. t .. ' '..self.filename
  osexecute(s)
  os.remove(t)
end

function scan:reg_f3d(_a)
  local a, target = self:prepareforreg(_a)
  
  local aff
  if target.cpp=='' then target.cpp = target.filename .. '.cpp.nii' end
  if target.aff~='' then aff = '-aff ' .. target.aff else aff='' end
  local s=[["]]..niftyreg..[[reg_f3d]]..exe..[[" ]]..gpu..[[-target ]].. a.filename..[[ -source ]].. self.filename .. [[ ]]..aff..[[ -cpp ]]..target.cpp
  osexecute(s)
  if target.aff~='' then
    os.remove(target.aff)
    target.aff = ''
  end  
  if target.invaff~='' then
    os.remove(target.invaff)
    target.invaff = ''
  end  
  if target.dvf~='' then 
    os.remove(target.dvf)
    target.dvf = ''
  end
 
  if target.cpp=='' then return end
  if target.invdvf=='' then target.invdvf = target.filename .. '.invdef.nii' end
  if target.dvf=='' then target.dvf = target.filename .. '.def.nii' end
  local s=[["]]..niftyreg..[[reg_transform]]..exe..[[" -ref ]].. a.filename .. [[ -cpp2def ]].. target.cpp .. [[ ]] .. target.invdvf
  osexecute(s)
  local s=[["]]..niftyreg..[[reg_transform]]..exe..[[" ]]..gpu..[[-ref ]].. a.filename .. [[ -invDef ]].. [[ ]] .. target.invdvf .. [[ ]] .. self.filename .. [[ ]] .. target.dvf
  osexecute(s)
  os.remove(target.cpp)
  target.cpp = ''
end

function scan:view()
  os.execute('"'..viewerexe..'" '..self.filename) 
end

function scan:viewdef()
  os.execute('"'..viewerexe..'" '..self.dvf) 
end

function scan:contains()
  return false
end

function scan:append(a)
  require('niftydicom')
  
  if self:isempty() then
    self:assign(a)
  else
    local r = a:as(self)
    append_nifty(self.filename, r.filename)
    r:free()
  end
end

function scan:slice(frame)
  require('niftydicom')
  local dim= self:dimensions()
  slice_nifty(self.filename, self.filename, frame)
  for i=1, dim[4] do
    self.dicomfiles[i] = self.dicomfiles[i+frame*dim[4]]
  end
  for i=dim[4]+1, dim[4]*dim[5] do
    self.dicomfiles[i] = nil
  end
  self.info.Frame = frame
end

function scan:slicefrom(a, frame)
  require('niftydicom')
  local dim= a:dimensions()
  slice_nifty(a.filename, self.filename, frame)
  for i=1, dim[4] do
    self.dicomfiles[i] = a.dicomfiles[i+frame*dim[4]]
  end
  for k, v in pairs(a.info) do
    self.info[k] = a.info[k]
  end
  self.info.Frame = frame
end

function scan:fullname()
  local names = {self.name}
  local a=self.owner
  while a do
    table.insert(names, a.name)
    a = a.owner
  end
  local s=names[#names], i
  for i=#names-1, 1, -1 do
    s = s .. '.' .. names[i]
  end
  return s
end

function scan:findconnection(a, path, prev)
  path = path or {}
  prev = prev or {}
  
  if self==a then 
    return a, path 
  end
  
  if #path>100 then 
    error 'scan path has cycles'
  end
  
  if (self.owner~=nil) and (self.owner==a.owner) then
    local s = {}

    s.invdvf = ''
    s.aff    = ''
    s.typ    = 'updown'
    s.name   = a:fullname()

    if (self.invdvf~='' or self.aff~='') and  (a.aff=='' and a.invdvf=='') then
      s.invdvf = self.invdvf
      s.aff    = self.aff
    elseif (a.invdvf~='' or a.aff~='') and  (self.aff=='' and self.invdvf=='') then
      s.invdvf = a.dvf
      s.aff    = a.invaff
    end

    s.source = self.filename
    s.target = a.filename
    table.insert(path, s)
    return res, p
  end
  if self.owner then
    for k,v in pairs(self.owner) do
      if (v~=self) and (v~=prev) then
        local s = {}
        s.invdvf = ''
        s.aff    = ''

        if (self.invdvf~='' or self.aff~='') and  (v.aff=='' and v.invdvf=='') then
          s.invdvf = self.invdvf
          s.aff    = self.aff
        elseif (v.invdvf~='' or v.aff~='') and  (self.aff=='' and self.invdvf=='') then
          s.invdvf = v.dvf
          s.aff    = v.invaff
        end
	
        s.typ        = 'updown'
        s.name       = v:fullname()
        s.source     = self.filename
        s.target     = v.filename
        table.insert(path, s)
        
        local res, p = v:findconnection(a, path, self)
        if res then 
          return res, p
        else
          table.remove(path, #path)
        end
     end
    end
  end
  return nil, nil
end

--/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
-- container object
--/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
-- a container is a table with extensions, all extensions must be listed before the definition of container:new
-- the purpose of container is to define the hierarchical relation of data, a container may contain a transform

container = {}
container.classname = 'container'
container.name      = ''
container.owner     = false
container.version   = ''
container.info      = {}

-- scan a folder (dir), find matching files and pass them to function f
function container:readfiles(dir, mat, f, max)
  local function scandir(directory)
    local i, t, popen = 0, {}, io.popen
    for filename in popen('dir '..directory..slashb):lines() do
        i = i + 1
        t[i] = filename
    end
    return t
  end
  local i,k,v; i=1;
  t = scandir(dir)
  for k,v in ipairs(t) do
    if string.find(v, mat)~=nil then
      local s = f(dir .. v)
      if s then
        self[i] = s
        i=i+1
        if max and i>max then
          break
        end
      end
    end
  end
end

-- scan a dicom server pass resulting objects to function f
-- this function only works in a DICOM server environment
function container:read_dicom(server, q, f, max, split4d)
  -- query for all scans, returning table of tables with series info or pass q as is (if already a table or dicomarray)
  local function scanquery(q)
    if type(q)=='userdata' then
      if #q==1 then
        if q.SeriesTime       ==nil then q.SeriesTime  ='' end
        if q.SeriesDate       ==nil then q.SeriesDate  ='' end
        if q.SeriesNumber     ==nil then q.SeriesNumber='' end
        if q.PatientID        ==nil then q.PatientID=''    end
        if q.Modality         ==nil then q.Modality=''     end
        if q.SeriesDescription==nil then q.SeriesDescription='' end
        if q.SeriesInstanceUID==nil then q.SeriesInstanceUID='' end
        q = dicomquery(server, 'SERIES', q)
        local t = {}
        for i=0,#q-1 do
          table.insert(t, {name=q[i].PatientID..':'..q[i].SeriesInstanceUID, 
                           SeriesTime=q[i].SeriesTime, SeriesDate=q[i].SeriesDate, SeriesNumber=q[i].SeriesNumber, 
                           SeriesDescription=q[i].SeriesDescription, Modality=q[i].Modality
                          })
        end
        table.sort(t, function(a, b) 
	                return (a.SeriesTime+0 < b.SeriesTime+0) or 
			((a.SeriesTime==b.SeriesTime) and (a.SeriesNumber+0<b.SeriesNumber+0)) 
		      end)
        q = {}
        for i=1,#t do 
          table.insert(q, t[i])
        end
      end
    end
    return q
  end

  -- this reader takes a patid:seriesuid string, table or dicom object as q
  local function defaultreader(server, q)
    local s = scan:new()
    local t = DicomObject:new()
    if type(q)=='string' then
      t.PatientID         = string.gsub(q, ':.*', '')
      t.SeriesInstanceUID = string.gsub(q, '.*:', '')
    elseif type(q)=='table' then
      t.PatientID         = string.gsub(q[1] or q.name, ':.*', '')
      t.SeriesInstanceUID = string.gsub(q[1] or q.name, '.*:', '')
    else
      t.PatientID = q.PatientID
      t.SeriesInstanceUID = q.SeriesInstanceUID
    end
    t.QueryRetrieveLevel = 'IMAGE'
    t.SOPInstanceUID     = ''
    t.ImageType          = ''
    local r = dicomquery(server, 'IMAGE', t)
    if #r<10 then
      print('ignoring data with less than 10 slices')
      return nil
    end
    if string.match(r[1].ImageType, 'DERIVED') then
      print('ignoring derived data')
      return nil
    end
    if type(q)=='table' then
      print('reading series', q.Modality, q.SeriesNumber, q.SeriesDescription)
    end
    s:read_dicom(server, q)
    if s:isempty() then 
      return nil
    end
    if type(q)=='table' then
      s.info.SeriesTime        = q.SeriesTime or ''
      s.info.SeriesDate        = q.SeriesDate or ''
      s.info.SeriesNumber      = q.SeriesNumber or ''
      s.info.SeriesDescription = q.SeriesDescription or ''
      s.info.Modality          = q.Modality or ''
-- add s.info.SeriesInstanceUID
        
      if s.info.SeriesDescription=='' then 
        s.info.SeriesDescription = s.dicomfiles[1].ImageType 
      end
    end
    return s
  end

  f = f or defaultreader
  local t
  local i=1
  t = scanquery(q)

  if type(t)=='userdata' then
    for k=0, #t-1 do
      local s = f(server, t[k])
      if s then
        local dim = s:dimensions()
        if dim[1]==4 and split4d==true then
          for i=0, dim[5] do
            self[i] = scan:new()
            self[i]:slicefrom(s, i)
            print('saved 4D item', i)
            i=i+1
          end
          s:free()
        else
          self[i] = s
          print('saved container item', i)
          i=i+1
        end
        if max and i>max then
          break
        end
      end
    end
  else
    -- t must be a table of tables
    for k, v in ipairs(t) do
      local s = f(server, t[k])
      if s then
        local dim = s:dimensions()
        if dim[1]==4 and split4d==true then
          for j=0, dim[5]-1 do
            self[i] = scan:new()
            self[i]:slicefrom(s, j)
            print('saved 4D item', i)
            i=i+1
          end
          s:free()
        else
          self[i] = s
          print('saved container item', i)
          i=i+1
        end
        
        if max and i>max then
          break
        end
      end
    end
  end
end

function container:write(file)
  local script =  'this.name = "'..self.name..'"\n' ..
                  'this.version = "1.0"\n'
  
  local reserved = {owner=1, classname=1, __index=1}
  
  -- recursively create script
  local function save(t, base)
    if type(t)=='table' then
      for k, v in pairs(t) do
        local s 
        
        if type(v)=='string' then 
          s = string.format('%q', v) 
        elseif type(v)=='table' then
          if v.classname then
            s = v.classname .. ':new()'
          else
            s = '{}' 
          end
        else 
          s = tostring(v) 
        end
      
        if reserved[k]==nil and type(v)~='function' then
          if type(k)=='number' then
            script = script .. string.format("%s[%d] = %s\n", base, k, s)
          else
            script = script .. string.format("%s.%s = %s\n", base, k, s)
          end
          
          if type(v)=='table' then 
            if type(k)=='number' then
              s = string.format('%s[%d]', base, k)
            else
              s = string.format('%s.%s', base, k)
            end
            save(getmetatable(v), s) 
            save(v, s)
          end
        end
      end
    end
  end
  
  save(self, 'this')
  
  script = script .. '\n'

  -- write header
  local f = io.open(file, 'wb')
  f:write(script)
  f:close()
end

function container:read(file)
  local f = io.open(file, 'rb')
  local s = f:read('*all') -- only default headerlength for now
  f:close()
  s = 'local this = ...\n' .. s
  loadstring(s)(self)
end

function container:process(f, out)
  for k, v in pairs(self) do
    f(v, out, k)
  end
end

-- this function only works if the IUP/CD environment is present
function container:plot(f, caption)
  if string.find(package.cpath, 'clibs')==nil then
    package.cpath = package.cpath .. ';./clibs/?51.dll;./clibs/?.dll'
  end
  require('iuplua')
  require 'iuplua_pplot'
  require( "iupluacontrols" )
  
  caption = caption or {}
  
  function pack(...)
    return arg
  end

  plot = iup.pplot{
    TITLE = caption[2] or self.name,
    MARGINBOTTOM="55",
    MARGINLEFT="55",
    AXS_XLABEL=caption[3] or "Index",
    AXS_YLABEL=caption[4] or "Value",
    LEGENDSHOW="NO",
    LEGENDPOS="TOPLEFT",
    AXS_XCROSSORIGIN="NO", 
    AXS_YCROSSORIGIN="NO"
  }
  
  local i, j, p;
  p = {}
  for k, v in ipairs(self) do
    p[k] = pack(f(v))
  end
 
  for j = 1, #p[1] do
    iup.PPlotBegin(plot, 0)
    for i=1, #p do
      iup.PPlotAdd(plot, i, p[i][j])
    end
    iup.PPlotEnd(plot)
  end

  local d = iup.dialog
            { iup.vbox{plot}, 
              size="400x300", title=caption[1] or self.classname
	    }
  d:show()
end

function container:free()
  for k, v in pairs(self) do 
    if v.free then v:free() end
  end
end

function container:fullname()
  local names = {self.name}
  local a=self.owner
  while a do
    table.insert(names, a.name)
    a = a.owner
  end
  local s=names[#names], i
  for i=#names-1, 1, -1 do
    s = s .. '.' .. names[i]
  end
  return s
end

-- container is a normal table augmented with registration etc in metatable
function container:new()
  local s = {}
  local m = {}
  setmetatable(s, m)
  
  -- load extensions into metatable
  for k,v in pairs(self) do
    m[k] = v
  end
  
  -- index unknown keys in metatable
  m.__index    = m
  
  -- writing into container is specialized here
  m.__newindex = 
  function(t,k,v) 
    -- write into existing metatable properties    
    if getmetatable(t)[k] then
      getmetatable(t)[k]=v
    else
      if v.owner~=false and v.owner~=nil then 
        error ('object already in container ' .. v.owner.name)
      end
       
      -- set owner and name of in objects added to the table
      if (getmetatable(v).owner~=nil) then rawset(getmetatable(v), 'owner', t) else rawset(v, 'owner', t) end
      if (getmetatable(v).name~=nil)  then rawset(getmetatable(v), 'name',  k) else rawset(v, 'name',  k) end
      
      -- write into table
      rawset(t,k,v) 
    end
  end
  
  return s
end
